import re,json
from utility import directory_creator
CFG = {
        'api_id':17203958,
        'api_hash':'82cefc4001e057c9d1488ab90e23d54f',
        'device_model':'Telegram Desktop 4.12.2',
        'system_version':'5.15.1',
        'app_version':'1.6.7',
        'number':None,
        'session':None,
        'lang_code':'en',
        'system_lang_code':'en',
        'proxy':None,
        'password_2fa':None,
    }

def get_number(txt):
    txt = str(txt)
    numbers =re.findall(r'\d+', txt)
    if numbers:
        sn = ''.join(numbers)
        sn = sn.lstrip('0')
        return int(sn)


import phonenumbers
from phonenumbers.phonenumberutil import region_code_for_country_code
from phonenumbers.phonenumberutil import region_code_for_number
import pycountry

def get_country_name(phone_number):
    try:
        phone_number = str(phone_number)
        if phone_number[0]!='+':phone_number=f"+{phone_number}"
        pn = phonenumbers.parse(phone_number)
        country = pycountry.countries.get(alpha_2=region_code_for_number(pn)).name
        country = country.replace(' ','_').replace(',','')
        return country
    except:return 'other'

def account_path_list(number, dtype='json', add_country=False):
    number = str(number)
    number = number.replace('+', '')
    out = ['accounts', dtype]
    if add_country:
        out += [get_country_name(number)]
    out += [number]
    return out

def path_for_tlg(number, dtype='json', add_country=False):
    number = clean_number(number)
    pl = account_path_list(number, dtype=dtype, add_country=add_country)
    number = pl[-1]
    out=[]
    last_path=''
    for i in pl:
        path= f'{last_path}/{i}'
        out.append(path)
        last_path=path
    return directory_creator(out)+f'/{number}.{dtype}'

def cfg_from_json(number,json_path=None):
    number= clean_number(number)
    if not json_path:
        json_path=path_for_tlg(number)
    try:
        with open(json_path, 'r') as f:
            cfg = json.loads(f.read())
    except:cfg={'number':number}
    return ready_cfg(**cfg)

def clean_number(number):
    number=str(number)
    number = number.replace('+', '')
    return number

def ready_cfg(number: str, **data):
    number = clean_number(number)
    cfg=CFG.copy()
    cfg['number']= number
    cfg['session']= path_for_tlg(number,data.get('dtype','session'))
    for i in cfg:
        if val:=data.get(i):
            cfg[i]=val
    return cfg

def save_json(tlg, json_path: str, p2fa=None):
    init_request = tlg._init_request
    device_model = init_request.device_model
    system_version = init_request.system_version
    app_version = init_request.app_version
    system_lang_code = init_request.system_lang_code
    lang_code = init_request.lang_code
    cfg=CFG.copy()
    cfg['api_id'] = tlg.api_id
    cfg['number'] = tlg.number
    cfg['api_hash'] = tlg.api_hash
    cfg['device_model'] = device_model
    cfg['system_version'] = system_version
    cfg['app_version'] = app_version
    cfg['system_lang_code'] = system_lang_code
    cfg['lang_code'] = lang_code
    cfg['password_2fa'] = p2fa

    with open(json_path, 'w') as f:
        f.write(json.dumps(cfg))










# from lib.admin import Admin
# from lib.user import User

# from bots.tlg.bot.util.packs import Start
# from bots.tlg.bot import TlgBot,Conversation
# from bots.tlg.bot.util.filteres import *

# def Admin(bot: TlgBot):
#     self = Conversation(Role.ADMIN.value)
#     self.menu_keys = [
#         ['managing', 'setting'],
#         ['statistics', 'support'],
#     ]
    # Start(self)
#     ##########################



from bots.tlg.bot.util.enum import *
from bots.tlg.bot import TlgBot
import os,sys


############################# Config ############################################
try:
    if sys.argv[1]=='test':
        os.environ["testMode"]="1"
except Exception as e:pass

token = ''
proxy =None
if os.environ.get("testMode","0")=="1":
    print('Run As TestMode')
    token = ''
    proxy = 'http://127.0.0.1:8889/'
bot_dir ='/root/bot_dir'
##########################################################################

bot = TlgBot(bot_dir, [1119223961], token, proxy=proxy, editors_id=[1119223961])

# bot.add_table('service_types', {'id': 0, 'title': '', 'del': 0})

bot.add_setting('off_support_pm','off_support_pm','texts',hide=1)
bot.add_setting('add_hi','سلام وقت تون بخیر','texts')

bot.init_db()

if __name__ == "__main__":
    bot.add(Admin(bot))
    bot.add(User(bot))



from time import sleep, time
from telethon.sync import TelegramClient
from telethon.sync import events, functions, errors
import telethon
import re
from utility.proxy import PROXY
import nest_asyncio
from random import choice
nest_asyncio.apply()
from telethon.sessions import Session
from telethon.network.connection import Connection
from asyncio.events import AbstractEventLoop
from telethon.tl.functions.account import UpdateProfileRequest
from utility.service import translator
from utility import Logger,directory_creator

class TlgApp(TelegramClient):
    def __init__(self: TelegramClient, number:str,session: str | Session, api_id: int, api_hash: str, *,  use_ipv6: bool = False, proxy=None, local_addr: str | tuple = None, timeout: int = 10, request_retries: int = 5, connection_retries: int = 5, retry_delay: int = 1, auto_reconnect: bool = True, sequential_updates: bool = False, flood_sleep_threshold: int = 60, raise_last_call_error: bool = False, device_model: str = None, system_version: str = None, app_version: str = None, lang_code: str = 'en', system_lang_code: str = 'en', loop: AbstractEventLoop = None, base_logger: str = None, receive_updates: bool = True, catch_up: bool = False, entity_cache_limit: int = 5000,auto_connect:bool=True,**kwargs):
        """
            proxy={'proxy_type':'http,socks5','ip','port','username','password'}
        """
        self.number = number
        self.log = Logger('TelegramClient')
        if proxy:
           proxy = PROXY(**proxy).make_proxy()
        super().__init__(
            session, api_id, api_hash,  use_ipv6=use_ipv6, proxy=proxy, local_addr=local_addr, timeout=timeout, request_retries=request_retries, connection_retries=connection_retries, retry_delay=retry_delay, auto_reconnect=auto_reconnect, sequential_updates=sequential_updates, flood_sleep_threshold=flood_sleep_threshold, raise_last_call_error=raise_last_call_error, device_model=device_model, system_version=system_version, app_version=app_version, lang_code=lang_code, system_lang_code=system_lang_code, loop=loop, base_logger=base_logger, receive_updates=receive_updates, catch_up=catch_up, entity_cache_limit=entity_cache_limit)
        
        if auto_connect:
            self.connect()
        self.d={}
        
    def do_connect(self):
        async def act():
            await self.connect()
            return True
        return self.loop.run_until_complete(act())        
    
    async def _get_status(self):
        try:
            res = await self.get_me()
        except:
            return False
        if res == None:
            return False
        return True
   
   
    def acc_status(self):
        return self.loop.run_until_complete(self._get_status())

    
    def do_login(self, first_step=False, **args):
        '''
            steps: 0=> auto
                   1=>send code 
                   2=> enter code :args= code
                   3=> enter 2fa   :args= p2fa
        '''
        
        step = self.d.get('step')
        if first_step:
            step = 0
        if step == 0:
            self.d['code'] = ''
            self.d['hash'] = ''
            self.d['p2fa'] = ''
            step = 1
        hash = args.get('hash', '')
        code = str(args.get('code', ''))
        p2fa = str(args.get('p2fa', ''))
        if step == 1:
            if self.acc_status():
                return 'session_is_active'
            out = self.loop.run_until_complete(self._send_code())
            if out[0]:
                self.d['step'] = 2
                self.d['hash'] = out[1]
                return ''
            return out[1]
        elif step == 2:
            if hash == '':
                hash = self.d.get('hash', '')
            out = self.loop.run_until_complete(self._sign_in(code, hash))
            if out[0]:
                self.d['step'] = 0
            if out[1] == 'need_pass2':
                self.d['step'] = 3
                self.d['code'] = code
                self.d['hash'] = hash
            return out[1]
        elif step == 3:
            if code == '':
                code = self.d.get('code', '')
            if hash == '':
                hash = self.d.get('hash', '')
            out = self.loop.run_until_complete(
                self._sign_in(code, hash, p2fa))
            if out[0]:
                self.d['step'] = 0
            return out[1]
        self.d['step'] = 0
        return False, 'wrong parameter in (do_login)'

    def login_loop(self):
        res = self.do_login(first_step=True)
        if res == 'session_is_active':
            # print(res)
            return
        inp = input('code:')
        while True:
            res = self.do_login(code=inp)
            if res == 'need_pass2':
                inp = input('pass2:')
                break
            elif res != '':
                # print(res)
                inp = input('code:')
            else:
                print('success login')
                return

        while True:
            res = self.do_login(p2fa=inp)
            if res != '':
                # print(res)
                inp = input('pass2:')
            else:
                print('success login')
                return
    
    def send_code(self):
        return self.loop.run_until_complete(self._send_code())

    async def _send_code(self):
        '''
            if have error session been removed
        '''
        try:
            if not await self.is_user_authorized():
                auth = await self.send_code_request(self.number)
                auth_hahs = auth.phone_code_hash
                return True, auth_hahs
            else:
                return False, '⚠️خطا در ارسال کد !'
        except errors.PhoneNumberBannedError:
            return False, f'⚠️خطا در ارسال کد به شماره {self.number} شماره شما از تلگرام مسدود شده است !'
        except errors.PhoneNumberInvalidError:
            return False, f'⚠️ شماره {self.number} اشتباه است.'
        except errors.FloodWaitError as e3:
            return False, f'⏳ شماره {self.number} از سمت تلگرام محدود شده است و تا {e3.seconds} ثانیه دیگر قابل ثبت نیست.'
        except errors.PhoneNumberOccupiedError:
            return False, '⚠️خطا در ارسال کد !'
        except errors.PhoneNumberUnoccupiedError:

            return False, '⚠️خطا در ارسال کد !'
        except ConnectionError:
            return False, '❌ارور در ارسال کد لطفا دوباره امتحان کنید.'
        except Exception as e:
            self.log(e)
            return False, '⚠️خطا در ارسال کد !'

    async def _sign_in(self, code, hash, p2fa=None):
        try:
            if not await self.is_user_authorized():
                await self.sign_in(phone=f'{self.number}', code=(code), phone_code_hash=hash)
                return True, ''
            else:
                return False, 'unknow_error'
        except errors.SessionPasswordNeededError as e:
            if p2fa:
                try:
                    await self.sign_in(password=str(p2fa))
                    return True, ''
                except Exception as e:
                    return False, "wrong_pass2"
            else:
                return False, "need_pass2"

        except errors.PhoneCodeExpiredError:

            return False, 'expied_code'
        except errors.PhoneCodeInvalidError:

            return False, 'wrong_code'
        except Exception as e:
            self.log(e)
            return False, 'unknow_error'

    async def _sign_up(self, code, hash):
        try:
            fname = 'fname'
            lname = 'lname'

            await self.sign_up(code=code, first_name=fname, last_name=lname, phone_code_hash=hash)

            return True, ''
        except errors.SessionPasswordNeededError:

            return False, "need_pass2"
        except errors.PhoneCodeExpiredError:

            return False, 'expied_code'
        except errors.PhoneCodeInvalidError:

            return False, 'wrong_code'
        except Exception as e:
            self.log(e)
            return False, 'unknow_error'

    def on_new_message(self,func,**kargs):
        '''
        func(self:TlgApp,event,**kargs)
        '''
        @self.on(event=events.NewMessage)
        @self.on(event=events.Album)
        async def action(event):
            try:
                func(self,event,**kargs)
            except Exception as e:
                print(e)

        self.run_until_disconnected()
        
    def get_all_dialogs(self):
        '''return groups entity'''
        async def act():
            ids = []
            try:
                async for dialog in self.iter_dialogs():
                    tmp = dialog.to_dict().get('entity', {})
                    ids += [tmp.to_dict()]
            except Exception as e:
                self.log(e)
            return ids
        return self.loop.run_until_complete(act())

    def just_this_sessions_is_active(self):
        async def act(self: TlgApp):
            res = False
            async with self as client:
                result = await client(functions.account.GetAuthorizationsRequest())
                current_exist = False
                for i in result.authorizations:
                    i = i.to_dict()
                    if i['current']:
                        current_exist = True
                res = len(result.authorizations) == 1 and current_exist
            return res
        return self.loop.run_until_complete(act(self))

    def do_change_or_set_2fa(self, new_p2fa: str,old_p2f=None):
        async def act():
            if old_p2f in ['',None]:
                await self.edit_2fa(new_password=new_p2fa,hint='2fa')
                return True
            else:
                if str(new_p2fa) != str(old_p2f):
                    await self.edit_2fa(str(old_p2f), new_password=new_p2fa)
                    return True
            return False

        return self.loop.run_until_complete(act())

    def change_info(self,  first_name=None, last_name=None, bio=None):
        async def act():
            await self(UpdateProfileRequest(first_name=first_name, last_name=last_name, about=bio))
            return True

        return self.loop.run_until_complete(act())
    
    def send_text(self,user,text):
        async def act():
            res = await self.send_message(user,text)
            return res
        return self.loop.run_until_complete(act())
    
    
    def get_last_message(self,entity):
        async def act():
            out = await self.get_messages(entity)
            return out[0].message
        return self.loop.run_until_complete(act())
    
    

    def get_account_status(self):
        '''
        return 'limit' , limit string time
        return 'ok' , ''
        return 'ban,''
        '''
        ban_pm='Unfortunately, some phone numbers may trigger a harsh response from our anti-spam systems'
        self.send_text('SpamBot','/start')
        sleep(2)
        last_message = self.get_last_message('SpamBot')
        if 'your account is now limited' in last_message:
            pattern = "your account is now limited until (.*?UTC)"
            stime = re.findall(pattern,last_message)[0]
            return 'limit' , stime
        elif 'Good news, no limits are currently applied to your account' in last_message:
            return 'ok',''
        elif ban_pm in last_message:
            return 'ban',''
                
    def download_by_chatId_and_fileId(self,chat_entity,file_id:int,out_path):
        async def act():
            message =await self.get_messages(chat_entity,ids=file_id)
            try:                
                if type(message.media)!=type(None):
                    out = await self.download_media(message,out_path)
                    return out
            except Exception as e:print(e)

        return self.loop.run_until_complete(act())
    
    def get_login_code(self):
        async def act():
            async for dialog in self.iter_dialogs():
                tmp = dialog.to_dict()
                if tmp.get('_') == 'Dialog' and tmp.get('name') == 'Telegram':
                    code = translator(dialog.message.message,'en')
                    # if 'Login code' in code:
                    find = re.findall('(\d{5}).',code)
                    if find:
                        return find[0],dialog.message.date
        return self.loop.run_until_complete(act())

    def logout(self):
        async def act():
            try:
                await self.log_out()
                return True
            except Exception as e:
                print(e)
                return False
        return self.loop.run_until_complete(act())
    
    def active_sessions(self):
        async def act():
            out=[]
            result = await self(functions.account.GetAuthorizationsRequest())
            for i in result.authorizations:
                    i = i.to_dict()
                    out+=[i]
            return out

        return self.loop.run_until_complete(act())
    
    def destroy_session(self,session_hash:int):
        async def act():
            try:
                await self(functions.account.ResetAuthorizationRequest(hash=int(session_hash)))
                return True
            except Exception as e:self.log(e)
            return False

        return self.loop.run_until_complete(act())

    def destroy_others_session(self):
        async def act():
            result = await self(functions.account.GetAuthorizationsRequest())
            for i in result.authorizations:
                i = i.to_dict()
                if not i['current']:
                    try:await self(functions.account.ResetAuthorizationRequest(hash=i['hash']))
                    except:pass

        return self.loop.run_until_complete(act())

    def seen_post(self, entity, ids:list[int]=[],limit=5):
        '''
            if ids==[] use limit
        '''
        async def _seen_post():
            try:
                if not ids:
                    list_of_id = []
                    async for msg in self.iter_messages(entity, limit=limit):
                        list_of_id += [int(msg.id)]
                else:list_of_id = ids
                await self(telethon.tl.functions.messages.GetMessagesViewsRequest(peer=entity, id=list_of_id, increment=True))
                return True
            except Exception as e:
                self.log(e)
            return False
            
        return self.loop.run_until_complete(_seen_post())
        
    def react_post(self, entity,reac:list, ids:list[int]=[],limit=5):
        '''
            if ids==[] use limit
            '''
        if not isinstance(reac,list):reac=[reac]
        async def _react_post():
            
            try:
                if not ids:
                    list_of_id = []
                    async for msg in self.iter_messages(entity, limit=limit):
                        list_of_id += [int(msg.id)]
                else:list_of_id = ids
                reaction=[telethon.tl.types.ReactionEmoji(
                    emoticon=choice(reac)
                )]
                for msg_id in ids:
                    await self(telethon.tl.functions.messages.SendReactionRequest(peer=entity,msg_id=msg_id,reaction=reaction))

                return True
            except Exception as e:
                self.log(e)
            return False
            
        return self.loop.run_until_complete(_react_post())
    
    def vote(self, entity, msg_id:int,option:str):
        async def act():
            try:
                await self(telethon.tl.functions.messages.SendVoteRequest(peer=entity,msg_id=msg_id,options=[option.encode()]))
                return True
            except Exception as e:
                self.log(e)
            return False
            
        return self.loop.run_until_complete(act())
    
    def join_channel(self, channel):
        async def _join_channel():
            try:
                res = await self(telethon.tl.functions.channels.JoinChannelRequest(channel))
                return True
            except Exception as e:
                self.log(e)
                return False
        return self.loop.run_until_complete(_join_channel())

    def leave_channel(self, channel):
        async def _leave_channel():
            try:
                await self(telethon.tl.functions.channels.LeaveChannelRequest(channel))
                return True
            except Exception as e:
                self.log(e)
                return False

        return self.loop.run_until_complete(_leave_channel())
    
    


import os,shutil
from .enum import *
from .filteres import *
from ..classes import Conversation
from telegram.ext import Filters
import time
from utility import directory_creator,root_path,ZipUtility

def setuper(value:str,subset:str,unit:str='',type:str='',hide:int=0,toggle:int=0):
    return {'value':value,'subset':subset,'unit':unit,'type':type,'hide':hide,'toggle':toggle}


def backup_database(self,chat_id=''):
    try:
        path = self.db.path
        database_dir = directory_creator(['database'])
        shutil.copy2(path,database_dir)
        zip_path = root_path()+'/database.zip'
        ziper=ZipUtility()
        ziper.zip_files(database_dir,zip_path)
        print(zip_path)
        with open(zip_path,'rb') as f:
            if chat_id=='':chat_id = self.host_id
            print(chat_id)
            self.bot.bot.send_document(chat_id, f,caption=self.get_bot_username())
    finally:
        try:os.remove(zip_path)
        except:pass
        try:shutil.rmtree(database_dir)
        except:pass

def get_product(self,id:int,status=CONFIRMED):
    try:
        return self.db.do('products',condition=f"id={id} AND status={status}")[0]
    except:return {}

def get_orders_by_pid(self,pid:int,status=WAITING):
    try:
        return self.db.do('orders',condition=f"pid={pid} AND status={status}")
    except:return []

def get_order(self,id:int,status=WAITING):
    try:
        return self.db.do('orders',condition=f"id={id} AND status={status}")[0]
    except:return {}

def setup_force_join(self):
    e = Node()
    e.filters = match_btn('force_join', self)

    def act(self: Conversation, *args):
        force_joins = self.db.do('force_join')
        if force_joins:
            self.send('lists', [['add_force_join'], [self.back_menu_key]])

            for i in force_joins:
                msg = f"{self.text('title',slash=True)}: {i['title']}\n"
                msg += f"{self.text('channel_id',slash=True)}: {i['channel_id']}"
                msg += f"{self.text('link',slash=True)}: {i['link']}"
                self.send(msg, {'delete_fj': i['id']}, t=False)
        else:
            self.send('no_item', [['add_force_join'], [self.back_menu_key]])
        return -1
    e.callback = act
    self.node('force_join', [e])
    ########################################
    e = Node()
    e.pattern = self.check_inline_keyboards('delete_fj')

    def act(self: Conversation, *args):
        id = int(self.splited_query_data()[1])
        self.db.do('force_join', delete=True)
        self.edit_message_text('deleted', self.get_msg_id())
        return -1
    e.callback = act
    self.node('delete_fj', [e])
    ########################################
    e = Node()
    e.filters = match_btn('add_force_join', self)
    e.msg = 'enter_title'
    e.btn = [[self.back_menu_key]]

    s = Node()
    s.filters = Filters.text
    s.msg = 'enter_channel_id'
    s.btn = [[self.back_menu_key]]

    s1 = Node()
    s1.filters = Filters.text
    s1.msg = 'enter_channel_link'
    s1.btn = [[self.back_menu_key]]

    s2 = Node()
    s2.filters = Filters.text
    def act(self: Conversation, *args):
        title = self.userdata(self.stat_key('add_force_join', 1))['text']
        channel_id = self.userdata(self.stat_key('add_force_join', 2))['text']
        link = self.get_text()
        self.db.do('force_join', {'id': self.time(), 'title': title,'channel_id':channel_id, 'link': link,'status':CONFIRMED})
        self.send('done', self.menu_keys)
        return -1

    s2.callback = act
    self.node('add_force_join', [e], [[s],[s1], [s2]])
 



from .time_functions import get_timestamp
def force_check_get_status_db(self,uid):
    try:
        if uid:
            user_item=self.db.do("users",condition=f"id={uid}")[0]
            force_check_ts=user_item['force_check_ts']
            force_check_staus=user_item['force_check_staus']
            if get_timestamp('now') - force_check_ts<5000:
                return force_check_staus
    except Exception as e:print(e)

def force_check_set_status_db(self,status,uid):
    if uid:
        self.db.do("users",{'force_check_ts':get_timestamp('now'),'force_check_staus':status},condition=f"id={uid}")
    
   



from telegram.utils.request import Request
from telegram.ext import Defaults, Updater, Filters
from telegram.ext import MessageHandler, ConversationHandler
import json
import os
from .util.filteres import RoleFilter
from .classes import Base, Conversation
from .util.enum import ConversationNode
from telegram import Bot, Update
from .util.functions import setuper
from utility import Logger, DB,root_path


class TlgBot(Bot):
    def __init__(self, BaseDirName: str, superAdmins: list[int], token: str, host_id: str = None, base_url: str = None, base_file_url: str = None, request: Request = None, private_key: bytes = None, private_key_password: bytes = None, defaults: Defaults = None, proxy: str = None, **kwargs):
        '''
            webhook_url: str = None,
            port: int = 443,
            ssl_key_path: str = None,
            ssl_cert_path: str = None,
            editors_id:list=[],
            respond_bad_order=True,
        '''
        os.environ['root_path'] = BaseDirName
        request_kwargs = None
        self.proxy=proxy
        if proxy:
            request_kwargs = {'proxy_url': proxy}
            request = Request(proxy_url=proxy)
        super(TlgBot, self).__init__(token, base_url, base_file_url,
                                     request, private_key, private_key_password, defaults)
        self.superAdmins: list = superAdmins
        self.webhook_inited=False
        self.host_id: list = host_id
        self.root = root_path()
        self.webhook_url: str = None
        self.port: int = 443
        self.ssl_key_path: str = None
        self.ssl_cert_path: str = None
        self.editors_id: list = []
        self.respond_bad_order = True
        self._set_attrs(**kwargs)
        self.log = Logger(BaseDirName).debug
        self.db: DB = None
        self._tables = {}
        self._settings = {}
        self._default_db_setups()
        self.updater = Updater(self.token, use_context=True,
                               request_kwargs=request_kwargs)
        self.dp = self.updater.dispatcher
        self.bot_username = None

    def _default_db_setups(self):
        self.add_table('users', {'id': 0, 'name': '', 'last_name': '', 'username': '', 'role': 0,'status': 0, 'phone': 0, 'language': 0, 'get_alarm': 0, 'is_online': 0, 'time': 0, 'last_seen': 0})
        self.add_table('setting', {'key': '', 'value': '', 'unit': '',
                       'type': '', 'subset': '', 'hide': 0, 'toggle': 0})
        self.add_table('texts', {'id': 0, 'key': '',
                       'value': '', 'language': 0, 'role': 0})
        self.add_table('media', {
                       'id': 0, 'key': '', 'msg_id': '', 'channel_id': '', 'language': 0, 'role': 0})
        self.add_setting('start_pm', 'سلام', subset='texts', hide=1)
        self.add_setting('bot_status', '1', type='int',
                         subset='status', hide=1)
        self.add_setting('token', self.token, subset='', hide=1)
        self.add_setting('host_id', self.host_id, subset='', hide=1)
        self.add_setting('super_admins', ','.join(
            [str(i) for i in self.superAdmins]), subset='', hide=1)
        self.add_setting('bot_username', 'Star1Vpn', subset='', hide=1)
        
    def _set_attrs(self, **kwargs):
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def _get_attr(self, key):
        return self.__dict__.get(key)

    def _init_settings(self):
        setting = self.db.do('setting')
        keys = [i['key'] for i in setting]
        for k, v in self._settings.items():
            if k not in keys:
                print(f"add setting key: {k}")
                self.db.do('setting', {**{'key': k}, **v})

        setting = self.db.do('setting')
        keys = [i['key'] for i in setting]
        must_be_remove = list(set(keys) - set(self._settings.keys()))
        for k in must_be_remove:
            print(f"remove setting key: {k}")
            self.db.do('setting', condition=f"key='{k}'", delete=True)

        self.settings_subset_update()

    def add_table(self, table_name: str, keys_values: dict):
        self._tables[table_name] = {**self._tables.get(table_name, {}), **keys_values}

    def add_setting(self, key: str, value: str, subset: str = '', unit: str = '', type: str = '', hide: int = 0, toggle: int = 0):
        self._settings[key] = setuper(
            value=value, subset=subset, unit=unit, type=type, hide=hide, toggle=toggle)

    def init_db(self):
        self.db = DB(self._tables, name=self.__dict__.get(
            'db_name', 'database'))
        self._init_settings()

    def settings_subset_update(self):
        setting_init = self._settings
        if setting_init:
            for k, v in setting_init.items():
                sub = v['subset']
                hide = v['hide']
                toggle = v['toggle']
                self.db.do('setting', {
                           'subset': sub, 'hide': hide, 'toggle': toggle}, condition=f"key='{k}'")

    def add(self, class_item: Base):

        if class_item.type == 'Conversation':
            item: Conversation = class_item
            for node in item.nodes:
                node: ConversationNode = node
                self.dp.add_handler(ConversationHandler(
                    entry_points=node.entries,
                    states=node.states,
                    fallbacks=node.fallbacks,
                    allow_reentry=node.arges.get('allow_reentry', True),
                    **node.arges
                ))
            if self._get_attr('respond_bad_order'):
                self.dp.add_handler(MessageHandler(
                    Filters.all & RoleFilter(item), item.add_method(item.not_fount)))

   

    def get_bot_username(self):
        self.bot_username = self.dp.bot.get_me().username
        return self.bot_username
        
    def befor_run_action(self):
        self.get_bot_username()
        self.db.do('setting',{'value':self.bot_username,'key':'bot_username'},condition="key='bot_username'")
        # self.deleteWebhook()
        print(self.bot_username)

    def init_webhook(self):
        self.befor_run_action()
        webhook_url = self._get_attr('webhook_url')
        port = self._get_attr('port')
        if port not in [443, 80]:
            webhook_url += f':{port}'
        webhook_url += f'/webhook/{self.token}'
        self.setWebhook(webhook_url, max_connections=100)

    def run_poll(self):
        self.befor_run_action()
        self.updater.start_polling(drop_pending_updates=False)
        self.updater.idle()

    def webhook(self, json_data):
        try:
            if not self.webhook_inited:
                self.webhook_inited=True
                self.init_webhook()
            data = json.loads(json_data)
            update = Update.de_json(data, self)
            self.dp.process_update(update)
        except Exception as e:
            print(e)



import requests,json
from time import sleep
from .util.enum import BTN_PFX, Role, HeadPack
from functools import wraps
import copy
from telegram.ext import MessageHandler, Filters, CommandHandler, CallbackQueryHandler
from .util.enum import ConversationNode, Node, Pack
from .util.filteres import IsReplied, match_btn, RoleFilter
from time import sleep
import telegram
from telegram.ext import Filters
from .util.filteres import match_btn
import types,json
from telegram import Update,MessageEntity,Bot,InlineKeyboardButton,InlineKeyboardMarkup,ReplyKeyboardMarkup,KeyboardButton
from utility import DB
from utility.structures import chunks
from .util.time_functions import get_timestamp
class Base:
    def __init__(self, super_self, *args, **kwargs) -> None:
        self.type = self.__class__.__name__
        self.super_self = super_self
        self.defulat_lang_code=kwargs.get('defulat_lang_code',0)
        self.languages=kwargs.get('languages',{})
        self.host_channel_id = super_self.host_id
        self.superAdmins = self.super_self.superAdmins
        self.update:Update = None
        self.context = None
        self.role = kwargs.get('role')
        self.user_info = None
        self.id = 0
        self.msg_id = 0
        self.back_key = 'back'
        self.back_menu_key = 'menu'
        self.menu_keys = []
        self.lang_code=-1
        self.text_list=[]
        self.log = super_self.log
        if type(self.log) == type(None):
            raise Exception('Debug Class Not defined')

        self.db: DB = super_self.db

        self.bot: Bot = super_self
        if type(self.bot) == type(None):
            raise Exception('bot Not defined')

        self.status = False
        self.checkpoints = {}

        self.go_menu_filter = match_btn(self.back_menu_key, self)
        self.go_back_filter = match_btn(self.back_key, self)

    def tail(self, func_name, *args, **kargs):
        if self.status:
            last_checkpoint = self.userdata('checkpoint')
            self.userdata('checkpoint', delete=True)
            if last_checkpoint:
                self.user_info = self.get_user_info_from_db([])
                self.run_checkpoint('NotNull')

    def head(self, func_name, *args, **kargs):
        print(func_name)
        if args:
            for i in args:
                if (i.__class__.__name__ == 'Update'):
                    self.update = i
                elif (i.__class__.__name__ == 'CallbackContext'):
                    self.context = i
                    
            try:
                if not self.update:
                    return
            except:
                return
            self.id = self.get_id()
            self.update_user()
            self.user_info = self.get_user_info_from_db([])
            if self.user('role') !=self.role or (self.user('status')!=1 and not self.is_super_admin()):
                self.status=False
                return
            if (self.user('role')!= Role.ADMIN.value and not self.is_super_admin()) and self.get_setting('bot_status')!=1:
                self.send('bot_off_text',get_text_setting=True)
                self.status=False
                return
            forbid_list = ['setter', 'menu_btns']
            print(func_name)
            for i in forbid_list:
                if i in func_name:break
            else:
                print(f"{func_name} set")
                self.userdata('last_function_name', func_name)
            self.status = self.run_checkpoint(func_name)

    def run_checkpoint(self, func_name):
        for key, func in self.checkpoints.items():
            res = func(self)
            go_on = True
            if not isinstance(res, bool):
                res, go_on = res
            if not go_on:
                self.userdata('checkpoint', key)
                return False

            

            if res in [False, None] :
                self.userdata('checkpoint', key)
                return False

            

        return True

    def detect_type(self, item: Node):
        if item.filters:
            return MessageHandler
        elif item.command:
            return CommandHandler
        elif item.pattern:
            return CallbackQueryHandler
        raise Exception(
            f"{self.element_name} havent true input for detect type")

    def dict_from_node(self, item: Node, idx: int, name: str, states_count: int = 0, level=-1):
        if not item.callback:
            def callback(self: Base, *args):
                id = self.stat_key(name,level)
                self.userdata(id,
                              {
                                  'text': self.get_text(),
                                  'query': self.splited_query_data(),
                                  'msg_id': self.get_msg_id(),
                              })
                msg = id
                if item.msg:
                    msg = item.msg
                btn = item.btn
                self.send(msg=msg, btns=btn,
                          chat_id=item.chat_id, **item.send_args)
                rtn_idx = -1
                if item.state_idx:
                    rtn_idx = item.state_idx

                elif level == (states_count) or level == -1:
                    rtn_idx = -1
                else:
                    rtn_idx = self.stat_key(name,level)
                return rtn_idx
            callback.__name__ = name+f"_{idx}_{level}"
            item.callback = self.add_method(callback, name+f"_{idx}")
        else:
            item.callback = self.add_method(item.callback,name+f"_{idx}")
        out = {'callback': item.callback}
        if item.filters:
            out['filters'] =  item.filters & RoleFilter(self)
            if level!=0:
                out['filters'] = ~self.go_menu_filter & out['filters']
        elif item.command:
            out['command'] = item.command
            
            out['filters'] = RoleFilter(self)
            if level!=0:
                out['filters'] = ~self.go_menu_filter & out['filters']
            if item.command_filters:
                out['filters'] = out['filters'] & item.command_filters

        elif item.pattern:
            if isinstance(item.pattern, str):
                item.pattern = self.check_inline_keyboards(item.pattern)
            out['pattern'] = item.pattern
            out['pass_user_data'] = True
            

        return out

    def menu(self,*args):
        self.send('menu_text',self.menu_keys)
        return -1
    
    def stat_key(self,name,level):
        out =  f"{name}_{level}_{str(self.role)}"
        return out
    
    def add_method(self, func, name=''):
        if name == '':
            name = 'mf_' + str(self.role) + f"_{str(func.__name__)}"
        self.__dict__[name] = types.MethodType(self.action_on_call(func,name), self)
        return self.__dict__[name]

    def node(self, name: str, entry: list[Node], states: list[Node] = None, fallbacks: list[Node] = None, db_config: dict[dict] = {}, **kargs):
        self.element_name = name
        if db_config:
            for k,v in db_config.items():
                self.super_self.add_table(k,v)
            self.super_self.init_db()
            
        conversation = ConversationNode()
        states_count = 0 if not states else len(states)
        level = 0
        for i, item in enumerate(entry):
            handler = self.detect_type(item)(
                **self.dict_from_node(item, idx=i, name=name, states_count=states_count, level=level))
            conversation.entries += [handler]
            level += 1
        # conversation.entries=[CommandHandler('start',self.start)]
        if states:
            for i, state in enumerate(states):
                if not isinstance(state, list):
                    state = [state]
                result = []
                for item in state:
                    handler = self.detect_type(item)(
                        **self.dict_from_node(item, idx=i, name=name, states_count=states_count, level=level))
                    result += [handler]
                idx= self.stat_key(name,level-1)
                conversation.states[idx] = result
                level += 1
        if fallbacks:
            for i, item in enumerate(fallbacks):
                conversation.fallbacks += [self.detect_type(item)(
                    **self.dict_from_node(item, idx=i, name=name))]
        else:
            conversation.fallbacks = [MessageHandler(
                match_btn(self.back_menu_key, self), self.add_method(self.menu))]
        conversation.arges = {**kargs}
        self.nodes += [conversation]

    def pack(self, pack, repalce_callback_func_name: str = None,**pack_kargs):
        if pack_kargs:
            pack: Pack = pack(self,**pack_kargs)
        else:pack: Pack = pack(self)
        if repalce_callback_func_name:
            for entry in pack.entry:
                try:
                    entry.callback.__name__ = repalce_callback_func_name
                except Exception as e:
                    self.log(e)

        self.node(
            name=pack.name,
            entry=pack.entry,
            states=pack.states,
            fallbacks=pack.fallbacks,
            db_config=pack.db_config,
            **pack.kargs
        )
        return pack

    def action_on_call(self, func,name):
        @wraps(func)
        def wrapper(self, *args, **kw):
            self.head(name, *args, **kw)
            res = None
            try:
                try:
                    status = self.status
                except:
                    status = False
                if status:
                    try:
                        res = func(self, *args, **kw)
                    except Exception as e:
                        self.log(e)
            finally:
                pass
            self.tail(func.__name__, *args, **kw)

            return res
        return wrapper

    def add_head(self, head_pack: HeadPack):
        '''
            func(self)
        '''
        head_pack = head_pack(self)
        if head_pack.db_config:
            self.super_self.db_config = {
                **self.super_self.db_config, **head_pack.db_config}
        if head_pack.pack:
            self.pack(head_pack.pack, repalce_callback_func_name=head_pack.name)

        self.checkpoints[head_pack.name] = head_pack.func

    #################################################
    def time(self):
        return get_timestamp('now')

    def is_super_admin(self):
        if self.id in self.superAdmins:
            return True
        return False

    def get_id(self):
        if 'callback_query' in self.update.to_dict():
            return self.update.callback_query.message.chat.id
        else:
            return self.update.message.from_user.id

    def get_name(self):
        if 'callback_query' in self.update.to_dict():
            return self.update.callback_query.message.chat.first_name
        else:
            return self.update.message.from_user.first_name

    def get_username(self):
        try:
            if 'callback_query' in self.update.to_dict():
                return self.update.callback_query.message.chat.username
            else:
                return self.update.message.from_user.username
        except:
            pass
        return None

    def get_last_name(self):
        if 'callback_query' in self.update.to_dict():
            return self.update.callback_query.message.chat.last_name
        else:
            return self.update.message.from_user.last_name

    def get_reply_markup(self, per_row_item=3,**kargs):
        reply_markup = self.update.callback_query.message.reply_markup.inline_keyboard
        res = {}
        for i in reply_markup:
            for btn in i:
                if 'callback_data' in btn.to_dict():
                    item = str(btn.callback_data)
                    sitem = item.split(':')
                    res[sitem[0]] = ":".join(sitem[1:])
                else:
                    res[btn.text] = str(btn.url)
        return self.inline_keyboard_maker(res, col=per_row_item,**kargs)

    def get_msg_id(self):
        if 'callback_query' in self.update.to_dict():
            return self.update.callback_query.message.message_id
        else:
            return self.update.message.message_id

    def get_chat_type(self):
        try:
            if 'callback_query' in self.update.to_dict():
                return self.update.callback_query.message.chat.type
            else:
                return self.update.message.chat.type
        except:
            pass

    def get_chat_id(self):
        if 'callback_query' in self.update.to_dict():
            return self.update.callback_query.message.chat.id
        else:
            return self.update.message.chat.id

    def get_type(self):
        try:
            if 'callback_query' in self.update.to_dict():
                return self.update.callback_query.message.chat.type
            else:
                return self.update.message.chat.type
        except:
            pass
    
    def get_msg_entities(self):
        return json.dumps([i.to_dict() for i in self.update.effective_message.entities])

    def msg_entities_json_entities_class(self,json_msg_entities):
        return [MessageEntity(**i) for i in json.loads(json_msg_entities)]

    def get_text(self, callback_deny=False):
        try:
            if not callback_deny and 'callback_query' in self.update.to_dict():
                return self.update.callback_query.message.text
        except:
            pass
        try:
            return self.update.message.text
        except:
            pass
    
    def get_btn_key_as_text(self,txt):
        k,_ = self.text(txt, True, 'value',slash=False)
        return k.replace(BTN_PFX, '')

    def splited_query_data(self, splitor=':'):
        try:
            data = self.update.callback_query.data
            return data.split(splitor)
        except:
            pass

    ###################################################
    def toggle(self, table: str, column: str, id, cond_col='id', switch=[0, 1]):
        if isinstance(id,str):
            cond= f"{cond_col}='{id}'"
        else:cond= f"{cond_col}={id}"
        res = self.db.do(table, condition=cond)
        if res:
            new_val = switch[-1]
            if res[0].get(column) == switch[-1]:
                new_val = switch[0]
            self.db.do(table, {column: new_val},cond)
            return new_val

    def userdata(self, key, value=None, delete=False, regex=False):
        if delete:
            try:
                self.context.user_data.pop(key)
            except:
                pass
            return
        if type(value) == type(None):
            if regex:
                out = {}
                for k, v in self.context.user_data.items():
                    if key in k:
                        out[k] = v
                return out
            return self.context.user_data.get(key)
        else:
            # if value != self.get_btn_value(self.back):
            self.context.user_data[key] = value
    ###################################################

    ###################################################
    def get_setting(self, key, return_row=False):
        try:
            row = self.db.do('setting', condition=f"key='{key}'")
            row = row[0]
            if return_row:
                return row
            if row['type'] == '':
                return row['value']
            else:
                try:
                    # if row['type'] == 'Language':

                    #     return Language.get_value_by_name(row['value'])
                    return eval(row['type'])(row['value'])
                except Exception as e:
                    self.log(e)
                    return None
        except:
            return None

    def set_setting(self, key, value):
        try:
            row = self.db.do('setting', condition=f"key='{key}'")
            if row:
                row = row[0]
                if row['type'] != '':
                    try:
                        value = eval(row['type'])(value)
                    except Exception as e:
                        self.log(e)
                        return False
            self.db.do(
                'setting', {'key': key, 'value': value}, condition=f"key='{key}'")
            return True
        except Exception as e:
                self.log(e)
        return False

    def user(self, key):
        try:
            return self.user_info[key]
        except:
            data = self.get_user_info_from_db([])
            if isinstance(data, list):
                self.user_info = data[0]
                return self.user_info.get(key)
                
            elif isinstance(data, dict):
                self.user_info = data
                return self.user_info.get(key)

    def update_user(self):
        try:
            data = {
                'id': self.id,
                'name': self.get_name(),
                'last_name': self.get_last_name(),
                'username': self.get_username(),
                'is_online': 1,
                'last_seen': self.time()
            }
            if self.is_super_admin():
                data['role'] =  Role.ADMIN.value
            user = self.db.do('users', condition=f"id={self.id}")
            try:
                user = user[0]
            except:
                user = {}

            if user.get('time') in [0, None]:
                first_cfg = {
                    'editor': 0,
                    'status': 1,
                    'get_alarm': 1,
                    'phone': 0,
                    'language': self.defulat_lang_code,
                }
                data = {**data, **first_cfg}
                data['time'] = self.time()
                role = Role.USER.value
                if self.is_super_admin():
                    role = Role.ADMIN.value
                data['role'] = role
            self.db.do('users', data, condition=f"id={self.id}")

        except Exception as e:
            self.log(e)

    def get_user_info_from_db(self, key_s, id=0):
        '''
        if str : return str
        elif list and empty return all info as dict
        else return requested info in list as dict
        '''
        if id == 0:
            id = self.id
        try:
            if isinstance(key_s, list):
                if id==0:return {}
                if key_s:
                    return {k: v for k, v in self.db.do('users', condition=f"id={id}")[0].items() if k in key_s}
                return self.db.do('users', condition=f"id={id}")[0]
            else:
                return self.db.do('users', condition=f"id={id}")[0][key_s]
        except:
            pass
        return ''

    def update_text_list(self,force=False):
        if force or not self.text_list:
            db_language = self.user('language')
            lang_cond=''
            self.lang_code=db_language
            if db_language==None or db_language==4:db_language=0
            if db_language!=None:
                lang_cond = f"language={db_language}"
            if lang_cond=='':lang_cond=None
            self.text_list = [i for i in self.db.do('texts', condition=lang_cond) if i['role']== self.role]
        return self.text_list
            
    
    def text(self, key, return_pair=False, search_by='key', return_id=False, slash=True, **kargs):
        '''
        return pair , target,id
        '''
        return key
        # main_word = kargs.get('main_word')

        # if key in ['',None]:return key
        # if kargs.get('get_text_setting',False) and (setting_value:=self.get_setting(key)):
        #     return setting_value
        # if not isinstance(key,str):key=str(key)
        
        # if search_by == 'key' and slash and key[0] != '/':
        #     key = f"/{key}"
        # target = 'key' if search_by == 'key' else 'value'
        # pair = 'value' if search_by == 'key' else 'key'
        
        # self.update_text_list(force=(self.lang_code!=self.user('language')))
        # texts = [i for i in self.text_list if i[target]==key]
        # if texts:
        #     text = texts[-1]
        #     if len(texts)>1 and main_word!=None:
        #         tmp = [i for i in texts if i[pair]==main_word]
        #         if tmp:
        #             text = tmp[0]

            
        #     out = [text[pair]]
        #     if return_pair:
        #         out += [text[target]]
        #     if return_id:
        #         out += [text['id']]
        #     if len(out) > 1:
        #         return out
        #     return out[0]

        # else:

        #     out = [key]
        #     if return_pair:
        #         out += [key]
        #     if return_id:
        #         out += [None]
        #     if len(out) > 1:
        #         return out
        #     return out[0]

    def media(self, key):
        if key == None:
            return -1, -1
        db_language = self.user('language')
        media = [i for i in self.db.do(
            'media', condition=f"language={db_language} AND key='{key}'") if i['role'] == self.role]
        if media:
            media = media[-1]
            return media['msg_id'], media['channel_id']
        return -1, -1

    ###################################################
    def get_replied_text(self):
        try:
            txt = self.update.message.reply_to_message.text
            if txt:
                return txt
            cap = self.update.message.reply_to_message.caption
            if cap:
                return cap
        except:
            pass
        return ''

    def is_replied(self):
        try:
            if self.update.message.reply_to_message != None:
                return True
        except Exception as e:
            self.log(e)
        return False

    def replace_btn_label(self, key,main_word=None):
        if isinstance(key, str):
            value, text_key = self.text(key, True,main_word=main_word,slash=False)
            if value != text_key and text_key.endswith(BTN_PFX):
                return value

            if not key.endswith(BTN_PFX):
                key += BTN_PFX
            value, text_key = self.text(key, True,main_word=main_word,slash=False)
            return value
        else:
            if not isinstance(key,list):key= list(key)
            for i in range(len(key)):
                res = self.replace_btn_label(key[i])
                if res != None:
                    key[i] = res

    def inline_keyboard_maker(self, inputs, **args):
        '''
            inputs ={key:data}

            col = args.get('col', 3)
            reverse:bool=False
            return_raw:bool=False
            add_label:bool=True
            close=True

        '''
        col = args.get('col', 3)
        reverse = args.get('reverse', False)
        add_label = args.get('add_label', True)
        return_raw = args.get('return_raw', False)
        close=args.get('close',True)
        btns = []
        for k, v in inputs.items():
            k=k.replace('*','')
            key = self.text(k,slash=False)
            if k.lower() == 'share':
                btns += [InlineKeyboardButton(key, switch_inline_query=v)]
            elif 'http://' in str(v) or 'https://' in str(v):
                btns += [InlineKeyboardButton(key, v)]
            else:
                val = f'{v}'
                btns += [InlineKeyboardButton(key,callback_data=val)]
        btns = chunks(btns, col, reverse)
        if close:
            btns += [[InlineKeyboardButton('close',callback_data='close')]]   
        if return_raw:
            return btns
        return InlineKeyboardMarkup(btns)

    def btn_maker(self, btns, **args):
        '''

        return_raw:bool=False
        share_phone:bool=False ** for keyboard
        resize:bool=True ** for keyboard
        col = args.get('col', 3) ** for inline
        reverse:bool=False ** for inline


        '''
        share_phone = args.get('share_phone', False)
        return_raw = args.get('return_raw', False)
        resize = args.get('resize', True)

        if btns:
            if not isinstance(btns, (ReplyKeyboardMarkup, InlineKeyboardMarkup)):
                if isinstance(btns, list):
                    btns_list = copy.deepcopy(btns)
                    self.replace_btn_label(btns_list)
                    if return_raw:
                        return btns_list
                    share_phone = args.get('share_phone', False)
                    if share_phone:
                        btns_list[0][0] = KeyboardButton(btns_list[0][0], True)
                    return ReplyKeyboardMarkup(btns_list, resize)
                if isinstance(btns, dict):
                    return self.inline_keyboard_maker(btns, **args)

            return btns
        return None

    def setter(self):
        if Filters.text(self.update):

            text = self.get_text()
            if text == 'delete_media':
                replied = self.userdata('last_function_name')
                values = [i for i in self.db.do(
                    'media', condition=f"key='{replied}' AND language='{self.user('language')}'") if i['role'] == self.role]
                if values:
                    for i in values:
                        self.db.do('media', condition=f"id={i['id']}", delete=True)
                    self.userdata('last_function_name', delete=True)
                    self.send_message('media_deleted')
                return

            replied = self.get_replied_text()

            if replied in ['', None]:
                return
            val, key, id = self.text(
                key=replied, return_pair=True, return_id=True,slash=False)
            if not id:
                key, val, id = self.text(
                    key=replied, search_by='value', return_pair=True, return_id=True,slash=False)

            if text == 'delete_translate':
                if id:
                    self.db.do('texts', condition=f"id={id}", delete=True)
                    self.send_message('text_translate_deleted')
                else:
                    self.send_message('text_dosnt_exists')
                return
            if id:
                self.db.do(
                    'texts', {'value': text}, condition=f"id={id}")
            else:
                to_db = {
                    'id': self.time(),
                    'key': replied,
                    'value': text,
                    'language': self.user('language'),
                }
                to_db['role'] = self.role
                self.db.do('texts', to_db)
            self.send_message('text_changed')

        elif Filters.photo(self.update) or Filters.video(self.update):
            replied = self.userdata('last_function_name')
            print(replied)
            msg_id = self.send_to_host(self.get_msg_id(), self.get_chat_id())
            values = [i for i in self.db.do(
                'media', condition=f"key='{replied}' AND language='{self.user('language')}'") if i['role'] == self.role]
            if msg_id < 0:
                return -1
            if values:
                key = replied
                print(key)
                id = values[0]['id']
                self.db.do('media', {
                    'key': key,
                    'msg_id': msg_id,
                    'channel_id': self.host_channel_id,
                    'language': self.user('language'),
                    'role': self.role
                }, condition=f"id={id}")
            else:
                cfg = {'id': self.time(), 'key': replied, 'msg_id': msg_id, 'channel_id': self.host_channel_id,
                       'language':  self.user('language')}
                cfg['role'] = self.role
                self.db.do('media', cfg)
            self.userdata('last_function_name', delete=True)

            self.send('media_changed')

        self.update_text_list(force=True)
    ###################################################
    def bot_kargs(self, **kargs):
        disable_web_page_preview = kargs.get('disable_web_page_preview')
        reply_to_message_id = kargs.get('reply_to_message_id')
        parse_mode = kargs.get('parse_mode')
        disable_notification = kargs.get('disable_notification')
        entities = kargs.get('entities')
        caption_entities = kargs.get('caption_entities')
        
        out = {}
        if disable_web_page_preview:
            out['disable_web_page_preview'] = disable_web_page_preview
        if reply_to_message_id:
            out['reply_to_message_id'] = reply_to_message_id
        if parse_mode:
            out['parse_mode'] = parse_mode
        if disable_notification:
            out['disable_notification'] = disable_notification
        if entities:
            out['entities'] = entities
        if caption_entities:
            out['caption_entities'] = caption_entities
        return out

    def send(self, msg, btns=None, chat_id: int = 0, **args):
        '''
            translat = True or t=True
            slash=True
            get_text_setting:bool

            col:int
            reply_to_message_id:int,
            parse_mode:PasrMode
            disable_notification

            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False
            close:bool =True

        '''

        btn = self.btn_maker(btns, **args)
        if chat_id == 0:
            chat_id = self.id
        try:
            msg_id, host = self.media(self.userdata('last_function_name'))
            if str(msg_id) != '-1':
                return self.copy_message(
                    from_chat_id=host,
                    message_id=msg_id,
                    btns=btn,
                    caption=msg,
                    **args
                )

        except Exception as e:
            self.log(e)
        try:
            return self.send_message(
                msg=msg,
                btns=btn,
                chat_id=chat_id,
                **args
            )

        except Exception as e:
            self.log(e)
        return False

    def send_message(self, msg, btns=None, chat_id: int = 0, **args):
        '''
            translat = True or t=True
            slash=True

            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        args['slash'] = args.get('slash', True)
        args['translat'] = args.get('translat', True) and args.get('t',True)
        args['disable_web_page_preview'] = args.get(
            'disable_web_page_preview', True)
        btn = self.btn_maker(btns, **args)
        if chat_id == 0:
            chat_id = self.id
        if args.get('translat'):
            msg = self.text(msg, **args)
        out = self.context.bot.send_message(
            chat_id=chat_id,
            text= msg,
            reply_markup=btn,
            **self.bot_kargs(**args)
        )
        return out['message_id']
    
    def send_media(self, media:bytes,media_type:str,caption=None, btns=None, chat_id: int = 0, **args):
        '''
            media_type:photo,audio,video,document
            translat = True or t=True
            slash=True

            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]

            

            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        args['slash'] = args.get('slash', True)
        args['translat'] = args.get('translat', True) and args.get('t',True)
        btn = self.btn_maker(btns, **args)
        if chat_id == 0:
            chat_id = self.id
        if args.get('translat'):
            caption = self.text(caption, **args)
        if media_type=="photo":
            out = self.context.bot.send_photo(
                photo = media,
                chat_id=chat_id,
                caption= caption,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
        elif media_type=="animation":
            out = self.context.bot.send_animation(
                animation = media,
                chat_id=chat_id,
                caption= caption,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
        elif media_type=="video":
            out = self.context.bot.send_video(
                video = media,
                chat_id=chat_id,
                caption= caption,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
        elif media_type=="audio":
            out = self.context.bot.send_audio(
                audio = media,
                chat_id=chat_id,
                caption= caption,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
        elif media_type=="document":
            out = self.context.bot.send_document(
                document = media,
                chat_id=chat_id,
                caption= caption,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
        return out['message_id']

    def send_animation(self, photo:bytes,caption=None, btns=None, chat_id: int = 0, **args):
        '''
            translat = True or t=True
            slash=True

            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        args['slash'] = args.get('slash', True)
        args['translat'] = args.get('translat', True) and args.get('t',True)
        btn = self.btn_maker(btns, **args)
        if chat_id == 0:
            chat_id = self.id
        if args.get('translat'):
            caption = self.text(caption, **args)
        out = self.context.bot.send_photo(
            photo = photo,
            chat_id=chat_id,
            caption= caption,
            reply_markup=btn,
            **self.bot_kargs(**args)
        )
        return out['message_id']
    
    def copy_message(self, from_chat_id, message_id, btns=None, chat_id=0, caption=None, **args):
        '''
            translat = True or t=True
            slash=True
            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        args['slash'] = args.get('slash', True)
        args['translat'] = args.get('translat', True) and args.get('t',True)
        btn = self.btn_maker(btns, **args)
        if chat_id == 0:
            chat_id = self.id
        if args.get('translat'):
            caption = self.text(caption, **args)
        out = self.context.bot.copy_message(
            chat_id=chat_id,
            from_chat_id=from_chat_id,
            message_id=message_id,
            caption=caption,
            reply_markup=btn,
            **self.bot_kargs(**args)
        )
        return out['message_id']

    def edit_message_text(self, new_msg, message_id, btns=None, chat_id: int = 0, **args):
        '''
            translat = True or t=True
            slash=True
            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        args['slash'] = args.get('slash', True)
        args['translat'] = args.get('translat', True) and args.get('t',True)
        try:
            btn = self.btn_maker(btns, **args)
            if chat_id == 0:
                chat_id = self.id
            if args.get('translat'):
                new_msg = self.text(new_msg, **args)
            out = self.context.bot.edit_message_text(
                chat_id=chat_id,
                message_id=message_id,
                text=new_msg,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
            return True
        except Exception as e:
            self.log(e)
        return False

    def edit_message_caption(self, new_caption, msg_id, btns=None, chat_id: int = 0, **args):
        '''
            translat = True or t=True
            slash=True

            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        args['slash'] = args.get('slash', True)
        args['translat'] = args.get('translat', True) and args.get('t',True)
        try:
            btn = self.btn_maker(btns, **args)
            if chat_id == 0:
                chat_id = self.id
            if args.get('translat'):
                new_caption = self.text(new_caption, **args)
            out = self.context.bot.edit_message_caption(
                chat_id=chat_id,
                message_id=msg_id,
                caption=new_caption,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
            return True
        except Exception as e:
            self.log(e)
        return False

    def edit_message_media(self, new_media, msg_id, btns=None, chat_id: int = 0, **args):
        '''
            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        try:
            btn = self.btn_maker(btns, **args)
            if chat_id == 0:
                chat_id = self.id
            out = self.context.bot.edit_message_media(
                chat_id=chat_id,
                message_id=msg_id,
                media=new_media,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
            return True
        except Exception as e:
            self.log(e)
        return False

    def edit_message_reply_markup(self, new_btns, msg_id, chat_id: int = 0, **args):
        '''
            col:int
            reply_to_message_id:int ,
            parse_mode:PasrMode
            disable_notification
            entites=[entity]
            caption_entities=[entity]


            disable_web_page_preview:bool=False
            share_phone:bool=False
            reverse_btn:bool=False

        '''
        try:
            btn = self.btn_maker(new_btns, **args)
            if chat_id == 0:
                chat_id = self.id

            out = self.context.bot.edit_message_reply_markup(
                chat_id=chat_id,
                message_id=msg_id,
                reply_markup=btn,
                **self.bot_kargs(**args)
            )
            return True
        except Exception as e:
            self.log(e)
        return False

    def delete_message(self, msg_id, id=0):
        try:
            if id == 0:
                id = self.id
            self.context.bot.delete_message(
                chat_id=id,
                message_id=msg_id)
            return True
        except Exception as e:
            self.log(e)
            return False

    def send_to_host(self, msg_id, from_chat):
        try:
            host_channel_id = self.host_channel_id
            return self.context.bot.copy_message(message_id=msg_id, chat_id=host_channel_id, from_chat_id=from_chat)['message_id']
        except:
            self.send_message('error_copy_to_host')
        return -1

    def alarm_to_admins(self, msg, btn=None):
        try:
            for id in self.superAdmins:
                self.send(msg=msg, btns=btn, chat_id=id)
            admins = self.db.do('users', condition=f"role={Role.ADMIN.value}")

            for admin in admins:
                try:
                    if str(admin['id']) not in  [str(i) for i in self.superAdmins] and admin['status'] == 1 and (admin['get_alarm'] == 1):
                        self.send(msg=msg, btns=btn, chat_id=admin['id'])
                except Exception as e:
                    self.log(e)
        except Exception as e:
            self.log(e)

    def send_to_users(self, msg_id, chat_id, user_ids:list=None,filter_ids:list=None):
        users = self.db.do('users', condition=f"get_alarm={1} AND is_online={1}")
        if user_ids:
            users = [i for i in users if i['id'] in user_ids]
        if filter_ids:
            users = [i for i in users if i['id'] not in filter_ids]

        bot = Bot(self.super_self.token)

        for user in users:
            try:
                sleep(1)
                bot.copy_message(
                    chat_id=user['id'], from_chat_id=chat_id, message_id=msg_id)
            except Exception as e:
                e = str(e)
                if 'Flood control exceeded.' in e:
                    sleep(40)
                    try:
                        bot.copy_message(
                            chat_id=user['id'], from_chat_id=chat_id, message_id=msg_id)
                    except:
                        if 'user is deactivated' in e:
                            self.db.do(
                                'users', {'is_online': 0}, condition=f"id={user['id']}")

                if 'user is deactivated' in e:
                    self.db.do('users', {'is_online': 0},
                               condition=f"id={user['id']}")


    def send_file(self, byte_file, chat_id=0,caption=None):
        try:
            if chat_id == 0:
                chat_id = self.id
            
            self.context.bot.send_document(
                chat_id, byte_file, caption=caption)
            return True
        except Exception as e:
            self.log(e)
        return False

    def get_doc_id(self):
        return self.update.message.document.file_id
    
    def get_real_file_name(self):
        return self.update.message.document.file_name

    def download_file(self, file_id,path:str, file_name='',set_real_fname=True):
        '''
        return path
        '''
        try:
            bot = self.context.bot
            bot = self.bot

            file_info = bot.get_file(file_id)
            if set_real_fname:fname = self.get_real_file_name()
            else:fname = file_info.file_path.split('/')[-1].split('\\')[-1]
            fname = file_name if file_name != '' else fname
            path = path.rstrip('/')+'/'
            path += f'{fname}'
            file_info.download(path)
            return path
        except Exception as e:
            self.log(e)
        return ''

    ###################################################
    def wait(self,btns=None,delete=False):
        msg_id = self.send('wait', btns=btns)
        sleep(1)
        if delete:self.delete_message(msg_id)
        return msg_id
        

    def pagination_maker(self, items: list, per_page=5, msg_list=False,revers_list=False):
        '''
        return msg, btn

        '''
        items = chunks(list(reversed(items)), per_page)
        self.userdata('pagination_items', items)
        self.userdata('pagination_iter', 0)
        self.userdata('pagination_revers_list', revers_list)
        
        try:
            msg = self.pagination_msg_maker(msg_list)
            btn = self.pagination_btn_maker()
        except:
            msg, btn = '', []
            if msg_list:
                msg = []
        return msg, btn

    def pagination_msg_maker(self, msg_list=False):
        items = self.userdata('pagination_items')
        idx = self.userdata('pagination_iter')
        if not self.userdata('pagination_revers_list'):
            item = list(reversed(items[idx]))
        else:item = list(items[idx])
        if msg_list:
            return item
        msg = ''
        for i in item:
            msg += i
        return msg

    def pagination_btn_maker(self):
        items = self.userdata('pagination_items')
        idx = self.userdata('pagination_iter')
        if len(items) <= 1:
            return None
        else:
            total_idx = len(items)-1
            if idx == 0:
                return {'next_page': 'pagination'}
            elif idx == total_idx:
                return {'back_page': 'pagination'}
            else:
                return {'back_page': 'pagination', 'next_page': 'pagination'}

    def set_uid_inFirstRequest(self):
        try:
            print(self.update)
            # if self.id not in [0,None]:return
            # proxy = None
            # if self.bot.proxy!=None:proxy={'https':self.bot.proxy}
                
            # url = f"https://api.telegram.org/bot{self.bot.token}/getUpdates"
            # print(url)
            # data = (requests.get(url,proxies=proxy).json())
            # print(data)
            # self.bot.dp.process_update(Update.de_json(data,self.bot))
        except Exception as e:
            print(e)
    
    def check_inline_keyboards(self, txt,role:int,regex=False, *args):
        def checker(*args):
            try:
                # self.set_uid_inFirstRequest()
                item = args[0]
                if isinstance(item, str):
                    if regex:
                        return txt in item and (role == self.role or role==0)
                    else:
                        out= txt in item.split(':') and (role == self.role or role==0)
                        return out
            except:
                pass
            return False
        return checker

    def force_join_inline_filter(self):
        def checker(*args):
            try:
                channels = self.db.do('force_join')
                if len(channels) > 0:
                    for ch in channels:
                        channel_id = ch['channel_id']
                        try:
                            if not self.is_channel_member(channel_id):
                                return True
                        except Exception as e:
                            print(e)
            except Exception as e:print(e)
            return False
        return checker

    def is_channel_member(self, channel_id, user_id=0):
        if user_id == 0:
            user_id = self.id
        if user_id in [None,0]:return True
        try:
            status = self.context.bot.getChatMember(channel_id, user_id)['status']
            if status in ['member', 'administrator', 'creator']:
                return True
        except Exception as e:
            print(e)
        return False

    def alert(self, raw_msg, id=0,popup=False):
        if id == 0:
            id = self.update.callback_query.id
        try:
            self.context.bot.answer_callback_query(
                callback_query_id=id, text=raw_msg, show_alert=popup)
        except Exception as e:
            self.log(e)



from .filteres import match_btn
from threading import Thread
from ..base import Base
from ..classes import Conversation
from .enum import Pack, Node
from .filteres import match_btn, match_text
from telegram.ext import Filters
from .enum import *
from ..classes import Conversation
import telegram
from telegram.ext import Filters
import time
from functools import partial
from .filteres import *
from utility.structures import chunks
from utility.service import translator


def Start(self: Conversation):

    def act(self: Conversation, *args):
        try:
            self.send(msg='start_pm', btns=self.menu_keys,
                      get_text_setting=True)
        except Exception as e:
            print(e)
        return -1

    p = Pack()
    p.name = 'start'

    e1 = Node()
    e1.command = 'start'
    e1.command_filters = match_text('/start')
    e1.callback = act

    e2 = Node()
    e2.pattern = self.check_inline_keyboards('/start', 1)
    e2.callback = act
    p.entry = [e1]
    return p


def ForceJoinPack(self: Conversation, **kargs):
    p = Pack()
    get_text_setting = kargs.get('get_text_setting', False)
    p.name = "force_join_pack"
    p.db_config = {'force_join': {'id': 0, 'title': '', 'channel_id': '', 'link': '',
                                  'status': 0}, 'users': {'force_check_ts': 0, 'force_check_staus': 0}}

    e = Node()
    e.filters = ~ForceJoinFilter(self)

    def act(self: Conversation, *a):
        try:
            start_bot = self.splited_query_data()[0]
            if start_bot == 'start_bot':
                self.alert(self.text('join_alert_pm'))
                return -1
        except:
            pass
        btns = {}
        for i in self.db.do('force_join', condition=f"status={CONFIRMED}"):
            print(i)
            link = i['link']
            btns[i['title']] = link
            txt = self.get_text()
            if txt and '/start' in txt:
                btns['start_bot'] = txt
            else:
                btns['start_bot'] = 'start_bot'
        self.send(msg='join_pm', btns=btns, col=1,
                  get_text_setting=get_text_setting, close=False)
        return -2
    e.callback = act

    e1 = Node()
    e1.pattern = self.force_join_inline_filter()
    e1.callback = act

    ch = Node()
    ch.pattern = self.check_inline_keyboards('start_bot', 1)

    def act(self: Conversation, *a):
        do_force_join_del_pm(self)
        return -2

    ch.callback = act

    p.entry = [e, e1, ch]
    return p


def do_force_join_del_pm(self: Conversation):
    try:
        self.splited_query_data()[0]
        self.delete_message(self.get_msg_id())
    except:
        pass


def SharePhone(self: Conversation, **kargs):
    p = Pack()
    p.name = 'share_phone'
    get_text_setting = kargs.get('get_text_setting', False)

    e1 = Node()
    e1.filters = SharePhoneFilter(self)

    def show_pm(self, *a):
        self.send('share_phone', [['share_phone']],
                  share_phone=True, get_text_setting=get_text_setting)
        return -1
    e1.callback = show_pm

    e = Node()
    e.filters = Filters.contact

    def act(self: Conversation, *args):
        contact = self.update.effective_message.contact
        phone = contact.phone_number
        self.db.do('users', {'id': self.id, 'phone': int(
            phone)}, condition=f"id={self.id}")
        self.user_info = self.get_user_info_from_db([])
        self.send_message('phone_shared', self.menu_keys)
        return -1
    e.callback = act
    p.entry = [e1, e]
    return p


def SelectLanguagePack(self: Conversation, **kargs):
    get_text_setting = kargs.get('get_text_setting', False)
    # next_check= kargs.get('next_check')
    p = Pack()
    p.name = 'select_language'
    e = Node()
    e.filters = SelectLanguageFilter(self)

    def act(self: Conversation, *a):
        btn = {k: f"{v}:select_language" for k, v in self.languages.items()}
        self.send(msg='Select_your_Language', btns=btn,
                  get_text_setting=get_text_setting, close=False)
        return -1
    e.callback = act

    e2 = Node()
    e2.command = 'Language'

    def act(self: Conversation, *a):
        btn = {k: f"{v}:select_language" for k, v in self.languages.items()}
        self.send(msg='Select_your_Language', btns=btn,
                  get_text_setting=get_text_setting, close=False)
        # if type(next_check)!=type(None):
        #     next_check(self)
        return -1
    e2.callback = act

    e1 = Node()
    e1.pattern = self.check_inline_keyboards('select_language', 1)

    def act(self: Conversation, *args):
        lang_id = int(self.splited_query_data()[1])
        self.db.do('users', {'id': self.id,
                   'language': lang_id}, condition=f"id={self.id}")
        self.delete_message(self.get_msg_id())
        self.user_info = self.get_user_info_from_db([])
        self.send('language_set', self.menu_keys,
                  get_text_setting=get_text_setting)

        return -1
    e1.callback = act
    p.entry = [e, e1, e2]
    return p


def Referral(self: Conversation):
    p = Pack()
    p.name = 'Referral'
    p.db_config = {'users': {'presenter': 0}}

    e1 = Node()
    e1.btn = self.menu_keys
    e1.filters = match_btn('referral', self)

    def referral_text(self: Base, *args):
        msg = self.text('referral_text', slash=True)+'\n\n'
        msg += f'https://t.me/{self.super_self.bot_username}?start=ref{self.id}\n\n'
        msg += f"{self.text('invited')} : {len(self.db.do('users', condition=f'presenter={self.id}'))}"

        self.send(msg=msg, btns=self.menu_keys, translat=False)
        return -1
    e1.callback = referral_text

    e2 = Node()
    e2.btn = self.menu_keys
    e2.command = 'start'
    e2.command_filters = Filters.regex('/start ref\d+')

    def referral_action(self, *args):
        try:
            presenter = int(self.get_text().split(' ')[-1].replace('ref', ''))
        except:
            try:
                presenter = int(self.splited_query_data()[
                                1].split(' ')[-1].replace('ref', ''))
            except:
                presenter = 0
        try:
            submited_presenter = int(self.user('presenter'))
        except:
            submited_presenter = 0
        if submited_presenter == 0 and presenter != self.id:
            self.db.do('users', {'presenter': presenter},
                       condition=f"id={self.id}")
            self.send('new_subset_text', chat_id=presenter)
        self.send('start_pm', self.menu_keys)
        return -1

    e2.callback = referral_action

    e3 = Node()
    e3.btn = self.menu_keys
    e3.pattern = self.check_inline_keyboards('/start ref', 2, regex=True)
    e3.callback = referral_action

    p.entry = [e1, e2, e3]
    return p


def ToggleSetting(self: Conversation, **kargs):
    p = Pack()
    key = kargs.get('key', 'toggle_setting')
    p.name = key

    def setting_msg_btn(self):
        setting = self.db.do('setting', condition=f"subset='{key}'")
        msg = ''
        btn = {}
        for i in setting:
            msg += f"{self.text(i['key'],slash=True)} : {self.text(i['value'])}\n"
            btn[i['key']] = key
        return msg, btn

    e = Node()
    e.filters = match_btn(key, self)

    def act(self: Conversation, *args):
        msg, btn = setting_msg_btn(self)
        self.send(msg, btn, translat=False)
        return self.stat_key(key, 0)
    e.callback = act

    s = Node()
    s.pattern = self.check_inline_keyboards(key, 1)

    def act(self: Conversation, *args):
        k = self.splited_query_data()[0]
        res = self.toggle('setting', 'value', k, 'key', ['0', '1'])
        self.alert(self.text(res))
        msg, btn = setting_msg_btn(self)
        self.edit_message_text(msg, self.get_msg_id(), btn, t=False)
        return self.stat_key(key, 0)
    s.callback = act

    p.entry = [e]
    p.states = [[s]]

    return p


def Status(self: Conversation, **kargs):
    p = Pack()
    subset = kargs.get('key', 'toggle_setting')
    column = kargs.get('column', 'subset')

    msg = kargs.get('msg', 'click_option')
    table = kargs.get('table', 'setting')
    value = kargs.get('value', 'value')
    key = kargs.get('target', 'key')
    p.name = subset

    def setting_msg_btn(self):
        setting = self.db.do('setting', condition=f"{column}='{subset}'")
        btn = []
        for item in setting:
            sbtn = [
                telegram.InlineKeyboardButton(
                    self.text(item[key]), callback_data=f'{item[key]}:{subset}slbl'),
                telegram.InlineKeyboardButton(
                    self.text(item[value]), callback_data=f'{item[key]}:{subset}set_stat'),
            ]
            btn += [sbtn]
        btn = telegram.InlineKeyboardMarkup(btn)
        return msg, btn
    e = Node()
    e.filters = match_btn(subset, self)

    def act(self: Conversation, *args):
        msg, btn = setting_msg_btn(self)
        self.send(msg, btn)
        return -1
    e.callback = act
    ###############################
    s = Node()
    s.pattern = self.check_inline_keyboards(subset+'slbl', 1)

    def act(self: Conversation, *a):
        name = self.splited_query_data()[0]
        self.alert(self.text(name))
        return -1
    s.callback = act
    ###############################
    s1 = Node()
    s1.pattern = self.check_inline_keyboards(subset+'set_stat', 1)

    def act(self: Conversation, *a):
        name = self.splited_query_data()[0]
        res = self.toggle(table, value, name, cond_col=key, switch=['0', '1'])
        _, btn = setting_msg_btn(self)
        self.edit_message_reply_markup(btn, self.get_msg_id(), col=4)
        return -1
    s1.callback = act
    p.entry = [e, s, s1]
    return p


def Setting(self: Conversation, **kargs):
    p = Pack()
    key = kargs.get('key', 'public_setting')
    p.name = key
    slash = kargs.get('slash', True)

    def setting_msg_btn(self):
        setting = self.db.do('setting', condition=f"subset='{key}'")
        msg = ''
        btn = []
        for i in setting:
            if i.get('hide') != 1:
                msg += f"{self.text(i['key'],slash=slash)} : \n{i['value']} {self.text(i['unit'],slash=True)}\n---\n"
            btn += [i['key']]
        if msg == '':
            msg = self.text('select_option', slash=True)
        return msg, btn

    e = Node()
    e.filters = match_btn(key, self)

    def act(self: Conversation, *args):
        msg, btn = setting_msg_btn(self)
        self.send(msg, chunks(btn, 2)+[[self.back_menu_key]], translat=False)
        return self.stat_key(key, 0)
    e.callback = act

    s = Node()
    s.filters = Filters.text

    def act(self: Conversation, *args):
        msg, btn = setting_msg_btn(self)
        k = self.get_btn_key_as_text(self.get_text())
        if k in btn:
            msg = 'enter_new_value'
            self.userdata('set_key', k)
            self.send(msg, [[self.back_menu_key]])
            return self.stat_key(key, 1)
        else:
            btn += [self.back_menu_key]
            self.send(msg, chunks(btn, 2), col=2, translat=False)
            return self.stat_key(key, 0)
    s.callback = act

    s1 = Node()
    s1.filters = Filters.text

    def act(self: Conversation, *args):
        k = self.userdata('set_key')
        if self.set_setting(k, self.get_text()):
            self.send('updated')
        else:
            self.send(self.text('bad_input', slash=True)+'\n' +
                      self.text('enter_new_value', slash=True))
            return self.stat_key(key, 1)

        msg, btn = setting_msg_btn(self)
        self.send(msg, chunks(btn, 2) +
                  [[self.back_menu_key]], col=2, translat=False)
        return self.stat_key(key, 0)
    s1.callback = act

    p.entry = [e]
    p.states = [[s], [s1]]

    return p


def SettingPro(self: Conversation, **kargs):
    p = Pack()
    key = kargs.get('key', 'setting_pro')
    p.name = key
    slash = kargs.get('slash', True)

    def setting_msg_btn(self):
        setting = self.db.do('setting', condition=f"subset='{key}'")
        msg = ''
        btn = []
        toggle_btn = []
        toggle_msg = 'click_option'
        for i in setting:
            if i.get('hide') != 1:
                if i.get('toggle') == 1:
                    sbtn = [
                        telegram.InlineKeyboardButton(
                            self.text(i['key']), callback_data=f"{i['key']}:{key}slbl"),
                        telegram.InlineKeyboardButton(
                            self.text(i['value']), callback_data=f"{i['key']}:{key}set_stat"),
                    ]
                    toggle_btn += [sbtn]
                else:
                    msg += f"{self.text(i['key'],slash=slash)} : \n{i['value']} {self.text(i['unit'],slash=True)}\n---\n"
                    btn += [i['key']]

        if msg == '':
            msg = self.text('select_option', slash=True)
        if toggle_btn:
            toggle_btn = telegram.InlineKeyboardMarkup(toggle_btn)

        return msg, btn, toggle_msg, toggle_btn

    e = Node()
    e.filters = match_btn(key, self)

    def act(self: Conversation, *args):
        msg, btn, toggle_msg, toggle_btn = setting_msg_btn(self)
        if toggle_btn:
            self.send(toggle_msg, toggle_btn, translat=False)
        if btn:
            self.send(msg, chunks(btn, 2) +
                      [[self.back_menu_key]], translat=False)
        return self.stat_key(key, 0)
    e.callback = act

    ###############################
    ss = Node()
    ss.pattern = self.check_inline_keyboards(key+'slbl', 1)

    def act1(self: Conversation, *a):
        name = self.splited_query_data()[0]
        self.alert(self.text(name))
        return self.stat_key(key, 0)
    ss.callback = act1
    ###############################
    ss1 = Node()
    ss1.pattern = self.check_inline_keyboards(key+'set_stat', 1)

    def act2(self: Conversation, *a):
        name = self.splited_query_data()[0]
        res = self.toggle('setting', 'value', name,
                          cond_col='key', switch=['0', '1'])
        _, _, _, toggle_btn = setting_msg_btn(self)
        self.edit_message_reply_markup(toggle_btn, self.get_msg_id(), col=4)
        return self.stat_key(key, 0)
    ss1.callback = act2

    s = Node()
    s.filters = Filters.text

    def act(self: Conversation, *args):
        msg, btn, _, _ = setting_msg_btn(self)
        k = self.get_btn_key_as_text(self.get_text())
        if k in btn:
            msg = 'enter_new_value'
            self.userdata('set_key', k)
            self.send(msg, [[self.back_menu_key]])
            return self.stat_key(key, 1)
        else:
            self.send(msg, chunks(btn, 2) +
                      [[self.back_menu_key]], col=2, translat=False)
            return self.stat_key(key, 0)
    s.callback = act

    s1 = Node()
    s1.filters = Filters.text

    def act(self: Conversation, *args):
        k = self.userdata('set_key')
        if self.set_setting(k, self.get_text()):
            self.send('updated')
        else:
            self.send(self.text('bad_input', slash=True)+'\n' +
                      self.text('enter_new_value', slash=True))
            return self.stat_key(key, 1)

        msg, btn, _, _ = setting_msg_btn(self)
        self.send(msg, chunks(btn, 2) +
                  [[self.back_menu_key]], col=2, translat=False)
        return self.stat_key(key, 0)
    s1.callback = act

    p.entry = [e]
    p.states = [[s, ss, ss1], [s1]]

    return p


def Products(self: Conversation, **kargs):
    p = Pack()
    row = kargs.get('row', 5)
    p.name = 'products'
    p.db_config = {
        'products': {
            'id': 0,
            'title': '',
            'price': 0.,
            'status': 0,
            'time': 0,
        }
    }
    e = Node()
    e.filters = match_btn('products', self)

    def act(self: Conversation, *args):
        products = self.db.do('products')
        main_btn = [['add_product'], [self.back_menu_key]]
        if products:
            self.send('products_list', btns=main_btn)
            msg = []
            for i in products:
                msg += [f"{i['title']}\n/p_{i['id']}\n\n"]
            msg, btn = self.pagination_maker(msg, row)
            self.send(msg=msg, btns=btn, t=False)

        else:
            self.send('no_item', main_btn)
        return -1
    e.callback = act
    p.entry = [e]
    return p


def Orders(self: Conversation, **kargs):
    p = Pack()
    row = kargs.get('row', 5)
    p.name = 'orders'
    p.db_config = {
        'orders': {
            'id': 0,
            'pid': '',
            'uid': '',
            'price': 0.,
            'status': 0,
            'time': 0,
        }
    }
    e = Node()
    e.filters = match_btn('orders', self)

    def act(self: Conversation, *args):
        orders = self.db.do('orders')
        main_btn = self.menu_keys
        if orders:
            self.send(msg, btns=main_btn, translat=False)
            msg = []
            for i in orders:
                msg += [f"/o_{i['id']}\n"]
            msg, btn = self.pagination_maker(msg, row)
            self.send(msg=msg, btns=btn, t=False)
        else:
            self.send('no_item', main_btn)
        return -1
    e.callback = act
    p.entry = [e]
    return p


def Multi(self: Conversation, name, items: list, check_sign='✅', table='setting', cond='key=', target='value', target_type=str, close=True):
    items = dict(zip(range(len(items)), items))
    p = Pack()
    p.name = name
    e = Node()
    e.filters = match_btn(name, self)

    def act(self: Conversation, *a):
        msg = f'{name}_txt'
        value = int(str(self.db.do(table, condition=cond)[0][target]))
        if value == None:
            value = 0
        btn = {}
        for k, v in items.items():
            sign = ''
            if value == k:
                sign = check_sign+' '
            btn[sign+self.text(v)] = f"{k}:multi_{name}"
        self.send(msg, btn, close=close, col=1)
        return -1
    e.callback = act
    # self.node('name',[e])
    #############################
    e1 = Node()
    e1.pattern = self.check_inline_keyboards(f'multi_{name}', 1)

    def act(self: Conversation, *a):
        tid = int(self.splited_query_data()[1])
        value = int(str(self.db.do(table, condition=cond)[0][target]))

        if tid == value:
            return -1

        try:
            ctid = target_type(tid)
        except:
            ctid = tid
        self.db.do(table, {target: ctid}, condition=cond)
        self.user_info = self.get_user_info_from_db([])
        value = int(str(self.db.do(table, condition=cond)[0][target]))
        if value == None:
            value = 0
        btn = {}
        for k, v in items.items():
            sign = ''
            if value == k:
                sign = '✅ '
            btn[sign+self.text(v)] = f"{k}:multi_{name}"
        self.edit_message_reply_markup(
            btn, self.get_msg_id(), col=1, close=close)
        return -1
    e1.callback = act
    p.entry = [e, e1]
    return p
    # self.node(f'sub{name}',[e])


# from .functions import translator

# 3
def translate(self: Conversation):

    e = Node()
    e.filters = match_btn('translate', self)

    def act(self: Conversation, *a):
        btn = {k: f'{v}:src_lang' for k, v in self.languages.items()}
        self.send('select_from', btn)
        return self.stat_key('translate', 0)
    e.callback = act

    s = Node()
    s.pattern = self.check_inline_keyboards('src_lang', 1)

    def act(self: Conversation, *a):
        print(2)
        self.userdata('src_lang', int(self.splited_query_data()[1]))
        btn = {k: f'{v}:trgt_lang' for k, v in self.languages.items()}
        self.edit_message_text('select_target', self.get_msg_id(), btn)
        return self.stat_key('translate', 1)
    s.callback = act

    s1 = Node()
    s1.pattern = self.check_inline_keyboards('trgt_lang', 1)

    def act(self: Conversation, *a):
        self.delete_message(self.get_msg_id())
        self.wait()
        trgt_lang = int(self.splited_query_data()[1])
        src_lang = self.userdata('src_lang')
        texts = self.db.do(
            'texts', condition=f"role={self.role} AND language={src_lang}")
        ex_texts = [i['key'] for i in self.db.do(
            'texts', condition=f"role={self.role} AND language={trgt_lang}")]
        texts = [i for i in texts if i['key'] not in ex_texts]
        out = []
        langs_pfx = {v: k for k, v in self.languages.items()}
        target = langs_pfx[trgt_lang]
        for i in texts:
            value = i['value']
            # do translat
            value = translator(value, target)
            if value == None:
                continue
            i['value'] = value
            i['language'] = trgt_lang
            i['role'] = self.role
            i['id'] = self.time()
            out += [i.copy()]
            time.sleep(0.01)
        self.db.do('texts', out)
        self.send('done')
        return -1
    s1.callback = act
    self.node('translate', [e], [[s], [s1]])


def SendToAllBasic(self: Conversation):
    p = Pack()
    p.name = 'send_to_all_basic'

    e = Node()
    e.filters = match_btn('send_to_all', self)
    e.msg = 'enter_your_message'
    e.btn = [[self.back_menu_key]]

    s = Node()
    s.filters = Filters.all

    def act(self: Conversation, *args):
        msg_id, chat_id = self.get_msg_id(), self.get_chat_id()
        Thread(target=partial(self.send_to_users,
               msg_id, chat_id, None, [self.id])).start()
        self.send_message('message_is_sending', self.menu_keys)
        return -1
    s.callback = act

    p.entry = [e]
    p.states = [[s]]
    return p

# ************** not completed********************


def SupportAdmin(self: Conversation):
    p = Pack()
    p.name = 'SupportAdmin'
    p.db_config = {'support': {'id': 0, 'u_id': 0, 'msg_id': 0,
                               'response_id': 0, 'responser_id': 0, 'extra': '', 'status': 0, 'time': 0}}
    e = Node()
    e.filters = match_btn('support', self)

    def show_message_from_user(self: Conversation, item):
        try:

            lnk = None
            u_username = self.user('username')
            if u_username != None:
                lnk = f'https://t.me/{u_username}'
            try:
                name = self.user('name')[:13]
            except:
                name = 'info'
            btn = {
                'send_reply': f"{item['id']}",
                'user_info': f"{item['u_id']}",
                'mark_read': f"{item['id']}",
                'show_response': f"{item['id']}",
                'perv_sup_pm': f"{item['id']}:{item['u_id']}",
                'next_sup_pm': f"{item['id']}:{item['u_id']}",
            }
            tbtn = btn.copy()
            tbtn[name] = f"shinfo:{item['id']}"
            if lnk:
                tbtn['pv'] = lnk

            try:
                self.copy_message(
                    from_chat_id=item['u_id'], message_id=item['msg_id'], chat_id=self.id, btns=tbtn)
            except:
                try:
                    btn['info'] = f"shinfo:{item['time']}"
                    if lnk:
                        btn['pv'] = lnk
                        self.copy_message(
                            from_chat_id=item['u_id'], message_id=item['msg_id'], chat_id=self.id, btns=tbtn)
                except:
                    self.send('unreadable_message' +
                              f"\n/user_{item['u_id']}", btn)
                    self.db.do('support', {'status': 1},
                               condition=f"id={item['id']}")
        except Exception as e:
            self.send_message('unreadable_message' +
                              f"\n/user_{item['u_id']}\nerror")
            self.db.do('support', {'status': 1}, condition=f"id={item['id']}")
            self.log(e)

    def unreaded_messages(self: Conversation, *args):
        unreaded_messages = self.db.do('support', condition=f"status=0")
        self.send('wait', self.menu_btn)
        try:
            if not unreaded_messages:
                self.send('no_item', self.menu_keys)
            else:
                for item in unreaded_messages:
                    self.show_message_from_user(item)
        except Exception as e:
            self.log(e)
        return -1
    e.callback = unreaded_messages
    p.entry = [e]
    return p



from .base import *

from time import time
class Conversation(Base):
    def __init__(self,role:Role,super_self,*args, **kwargs) -> None:
        super().__init__(super_self,role=role,*args, **kwargs)
        self.element_name=None
        self.nodes:list[ConversationNode] =[]
        self.init_setter()
        self.init_pagination()
        self.init_menu()
        self.init_closer()

    def closer(self,*args):
        self.delete_message(self.get_msg_id())
        return -1
    
    def not_fount(self,*args):
        # not_fount_status = self.get_setting('not_fount_status')
        # if not_fount_status and not_fount_status==1:
        self.send('not_fount',self.menu_keys)
        return -1
    # ############### init_setter###################
    # def init_setter(self):
    #     e = Node()
    #     e.filters = ~Filters.contact & IsReplied()
    #     e.callback = self.setter_entry
    #     self.node('setter',entry=[e])

    # def setter_entry(self, *args):
    #     if (self.id in self.super_self.editors_id):
    #         try:
    #             self.setter()
    #         except Exception as e:
    #             self.log(e)
    #         return  -1

    # ############### init_pagination###################

    def init_pagination(self):
        e1 = Node()
        e1.pattern = self.check_inline_keyboards('next_page:pagination',0,regex=True)
        e1.callback = self.pagination_entry

        e2 = Node()
        e2.pattern = self.check_inline_keyboards('back_page:pagination',0,regex=True)
        e2.callback = self.pagination_entry
        self.node('setter',entry=[e1,e2])

    def pagination_entry(self, *args):
        order = self.splited_query_data()[0]
        if 'pagination_iter' in self.context.user_data:
            if order == 'next_page':
                self.userdata('pagination_iter',
                              self.userdata('pagination_iter')+1)
            elif order == 'back_page':
                self.userdata('pagination_iter',
                              self.userdata('pagination_iter')-1)
            try:
                msg = self.pagination_msg_maker()
                btn = self.pagination_btn_maker()
                self.edit_message_text(new_msg=msg,message_id=self.get_msg_id(),btns=btn,chat_id=self.id,translat=False)
            except Exception as e:
                self.log(e)
                self.delete_message(self.get_msg_id())
        else:
            self.delete_message(self.get_msg_id())
        return -1
    ############### init_menu###################
    def init_menu(self):
        e = Node()
        e.filters = match_btn(self.back_menu_key,self)
        e.callback =self.menu
        self.node('menu',entry=[e])
    
    def init_closer(self):
        e = Node()
        e.pattern = self.check_inline_keyboards('close',0)
        e.callback =self.closer
        self.node('closer',entry=[e])

    
